import json
import sys
from typing import Any, Dict, TextIO
from valohai.config import is_valohai_deployment
from valohai.internals.notebooks import is_in_notebook

_supported_types = [int, float]


class Logger:
    partial_logs: Dict[str, Any]

    def __init__(self) -> None:
        self.partial_logs = {}

    def __enter__(self) -> "Logger":
        self.partial_logs = {}
        return self

    def __exit__(self, type, value, traceback) -> None:  # type: ignore[no-untyped-def]
        self.flush()

    def log(self, *args, **kwargs) -> None:  # type: ignore[no-untyped-def]
        """Log a single name/value pair to be flushed into standard output later as batch.

        For a repeating iteration like a machine learning training loop, Valohai expects
        all logged values to be printed as a batch.

        Example:
            for epoch in range(10):
                with valohai.logger() as logger:
                    logger.log("epoch", epoch)
                    logger.log("accuracy", 0.54)
                    logger.log("loss", 0.123)

        Example 2:
            for epoch in range(10):
                logger = valohai.logger():
                logger.log("epoch", epoch)
                logger.log("accuracy", 0.54)
                logger.log("loss", 0.123)
                logger.flush_logs()

        Example 3:
            for epoch in range(10):
                with valohai.logger() as logger:
                    logger.log("epoch", epoch)
                    logger.log(acc=0.54, loss=0.123)

        All three examples will act exactly the same.

        :param name: Name of the variable being logged (example: learning_rate)
        :param value: Value of the logged variable

        """
        if len(args) % 2 != 0:
            raise ValueError(f"Odd number of arguments in {args} for log()")
        for key, value in zip(args[::2], args[1::2]):
            self._serialize(key, value)
        for key, value in kwargs.items():
            self._serialize(key, value)

    def flush(self) -> None:
        """Flush all the partial logs into standard as a batch.

        For a repeating iteration like a machine learning training loop, Valohai expects
        all logged values to be printed as single batch.

        Example:
            logger = valohai.logger():
            logger.log("epoch", epoch)
            logger.log("accuracy", 0.54)
            logger.log("loss", 0.123)
            logger.flush_logs()

        This will log all three metrics at once.
        """
        if self.partial_logs:
            to_print = self.partial_logs
            if is_valohai_deployment():
                # Wrap in `vh_metadata` so deployment log machinery detects this
                to_print = {"vh_metadata": to_print}
            # Start with \n, ensuring JSON prints on its own line
            print(  # noqa: T201
                f"\n{json.dumps(to_print, default=str)}",
                file=self._get_output_stream(),
            )
            self.partial_logs.clear()

    def _serialize(self, name: str, value: Any) -> None:
        self.partial_logs.update({str(name): value})

    def _get_output_stream(self) -> TextIO:
        """Get the output stream to print metadata logs to."""
        if is_in_notebook():
            # When in a notebook, Jupyter will have redirected stdout to the stream that will show up
            # in the notebook.  We need to print to the original stdout instead so Valohai machinery
            # can pick up the logs.
            return getattr(sys, "__stdout__", sys.stdout)
        return sys.stdout


logger = Logger



import os


def is_running_in_valohai() -> bool:
    return bool(os.environ.get("VH_JOB_ID"))


def is_valohai_deployment() -> bool:
    return bool(
        os.environ.get("VALOHAI_PORT")
        or os.path.exists("/valohai/valohai-metadata.json")
    )


def is_flat_output_path() -> bool:
    return bool(os.environ.get("VH_FLAT_LOCAL_OUTPUTS"))



def uri_to_filename(uri: str) -> str:
    return uri.rpartition("/")[-1].split("?", 1)[0]


def string_to_bool(value: str) -> bool:
    if value.lower() == "false":
        return False
    return True


def get_sha256_hash(filepath: str) -> str:
    try:
        from valohai_cli.utils.hashing import get_fp_sha256  # type: ignore
    except ImportError as ie:
        raise RuntimeError(
            "The `valohai-cli` module must be available for verifying hash"
        ) from ie

    with open(filepath, "rb") as f:
        return get_fp_sha256(f)  # type: ignore



import contextlib
import os
import tempfile
from typing import Any, Dict

from requests import Response
from valohai.internals.utils import uri_to_filename, get_sha256_hash


def resolve_datum(datum_id: str) -> Dict[str, Any]:
    try:
        from valohai_cli.api import request  # type: ignore
    except ImportError as ie:
        raise RuntimeError("Can't resolve datum without valohai-cli") from ie
    resp: Response = request(url=f"/api/v0/data/{datum_id}", method="GET")
    resp.raise_for_status()
    data = resp.json()
    assert isinstance(data, dict)
    return data


def verify_datum(datum_obj: Dict[str, Any], input_folder_path: str) -> str:
    filename = datum_obj["name"]
    file_path = os.path.join(input_folder_path, filename)
    if os.path.exists(file_path) and datum_obj["sha256"] == get_sha256_hash(file_path):
        return file_path
    raise Exception(
        f"The local file {file_path!r} does not exist, "
        f"or its checksum does not match with datum {datum_obj['id']}."
    )


# TODO: This is close to valohai-local-run. Possibility to merge.
def download_url(url: str, path: str, force_download: bool = False) -> str:
    if not os.path.isfile(path) or force_download:
        if url.startswith("datum://"):
            input_folder_path = os.path.dirname(path)
            response = resolve_datum(uri_to_filename(url))
            path = verify_datum(response, input_folder_path)
        else:
            _do_download(url, path)
    else:
        print(f"Using cached {path}")  # noqa

    return path


def _do_download(url: str, path: str) -> None:
    try:
        import requests
    except ImportError as ie:
        raise RuntimeError(
            f"The `requests` module must be available for download support (attempting to download {url})"
        ) from ie

    tmp_path = tempfile.NamedTemporaryFile().name
    print(f"Downloading {url} -> {path}")  # noqa
    r = requests.get(url, stream=True)
    r.raise_for_status()
    total = int(r.headers["content-length"]) if "content-length" in r.headers else None
    try:
        from tqdm import tqdm

        prog = tqdm(total=total, unit="iB", unit_scale=True)
    except ImportError:
        prog = contextlib.nullcontext()

    with prog as prog, open(tmp_path, "wb") as f:
        for chunk in r.iter_content(1048576):
            if prog:
                prog.update(len(chunk))
            f.write(chunk)
    os.replace(tmp_path, path)
